from functools import reduce

def checkAllItemsAreAST(*z):
    return reduce(lambda r, x: r and (isinstance(x, AST)), z, True)

def checkAllItemsAreValue(*z):
    return reduce(lambda r, x: r and (isinstance(x, Value)), z, True)

class AST:
    def __repr__(self):
        return "{}()".format(self.__class__.__name__)

    def describe(self):
        return self.__repr__()

    def __hash__(self):
        return hash(self.__repr__())

    def getVariables(self):
        return set()

    def getConditionVariables(self, is_parent_condition=False):
        return set()

class Variable(AST):
    def __init__(self, name, sort):
        assert isinstance(name, str)
        assert isinstance(sort, Sort)
        self.name = name
        self.sort = sort

    def __repr__(self):
        return self.name
        # return self.describe()

    def describe(self):
        return "{}(name={}, sort={})".format(self.__class__.__name__, self.name, self.sort)

    def __hash__(self):
        return hash(self.__repr__())

    def __eq__(a, b):
        return a.name == b.name # and a.sort == b.sort

    def getVariables(self):
        return set([self])

    def getConditionVariables(self, is_parent_condition=False):
        if is_parent_condition:
            return self.getVariables()
        else:
            return set()

    def to_smt2(self):
        return self.name
    
class Sort(AST):
    def __init__(self, name, values):
        assert isinstance(name, str)
        assert isinstance(values, set) or isinstance(values, list)
        assert checkAllItemsAreValue(*values)
        self.name = name
        self.values = set(values)

    def __repr__(self):
        return "{}({})".format(self.__class__.__name__, self.name)

    def __hash__(self):
        return hash(self.__repr__())

    def __eq__(a, b):
        return a.name == b.name or a.values == b.values

    def to_smt2(self):
        return self.name

class Value(AST):
    def __init__(self, name):
        assert isinstance(name, str)
        self.name = name

    def __repr__(self):
        return self.name

    def describe(self):
        return "{}({})".format(self.__class__.__name__, self.name)

    def __hash__(self):
        return hash(self.__repr__())

    def __eq__(a, b):
        return a.name == b.name

    def to_smt2(self):
        return self.name

### Used to describe equality x == y |-> x = Ref(y)
class Ref(AST):
    def __init__(self, var):
        assert isinstance(var, Variable)
        self.variable = var
    
    def __repr__(self):
        return "{}({})".format(self.__class__.__name__, self.variable)

    def __eq__(a, b):
        return  a.__class__.__name__ == b.__class__.__name__ and a.variable == b.variable

class Const(AST):
    def __init__(self):
        pass

    def __repr__(self):
        return "{}()".format(self.__class__.__name__)

    def __eq__(a, b):
        return a.__class__.__name__ == b.__class__.__name__

    def getVariables(self):
        return set()

    def to_smt2(self):
        return self.__class__.__name__

class NOp(AST):
    def __init__(self, *v):
        assert checkAllItemsAreAST(*v), "v={}".format(v)
        self.v = list(v)

    def __repr__(self):
        return "{}({})".format(self.__class__.__name__, ', '.join(['{}'.format(x) for x in self.v]))

    def __eq__(a, b):
        return a.v == b.v

    def getVariables(self):
        return reduce(lambda r, expr: r | expr.getVariables(), self.v, set())

    def getConditionVariables(self, is_parent_condition=False):
        return reduce(lambda r, expr: r | expr.getConditionVariables(is_parent_condition), self.v, set())

    def to_smt2(self):
        return "({} {})".format(self.__class__.__name__.lower(), ' '.join([x.to_smt2() for x in self.v]))

class UniOp(NOp):
    def __init__(self, v1):
        assert isinstance(v1, AST)
        self.v1 = v1
    
    def __repr__(self):
        return "{}({})".format(self.__class__.__name__, self.v1)

    def __eq__(a, b):
        return a.v1 == b.v1

    def getVariables(self):
        return self.v1.getVariables()

    def getConditionVariables(self, is_parent_condition=False):
        return self.v1.getConditionVariables(is_parent_condition)

    def to_smt2(self):
        return "({} {})".format(self.__class__.__name__.lower(), self.v1.to_smt2())

class BinOp(NOp):
    def __init__(self, v1, v2):
        assert isinstance(v1, AST)
        assert isinstance(v2, AST)
        self.v1 = v1
        self.v2 = v2
    
    def __repr__(self):
        return "{}({}, {})".format(self.__class__.__name__, self.v1, self.v2)

    def __eq__(a, b):
        return a.v1 == b.v1 and a.v2 == b.v2

    def getVariables(self):
        return self.v1.getVariables() | self.v2.getVariables()

    def getConditionVariables(self, is_parent_condition=False):
        return self.v1.getConditionVariables(is_parent_condition) | self.v2.getConditionVariables(is_parent_condition)

    def to_smt2(self):
        return "({} {} {})".format(self.__class__.__name__.lower(), self.v1, self.v2)

### Holds no AST nodes
class Terminate(Const):
    pass

### True
class Top(Const):
    def to_smt2(self):
        return "true"

### False
class Bot(Const):
    def to_smt2(self):
        return "false"

class Not(UniOp):
    pass

class Eq(BinOp):
    def to_smt2(self):
        return "(= {} {})".format(self.v1, self.v2)

class And(NOp):
    pass

class Or(NOp):
    pass

class If(AST):
    def __init__(self, cond_clause, then_clause, else_clause):
        assert isinstance(cond_clause, AST)
        assert isinstance(then_clause, AST)
        assert isinstance(else_clause, AST)
        self.cond_clause = cond_clause
        self.then_clause = then_clause
        self.else_clause = else_clause

    def __repr__(self):
        return "{}({}, {}, {})".format(self.__class__.__name__, self.cond_clause, self.then_clause, self.else_clause)

    def getVariables(self):
        return self.cond_clause.getVariables() | self.then_clause.getVariables() | self.else_clause.getVariables()

    def getConditionVariables(self, is_parent_condition=False):
        return self.cond_clause.getConditionVariables(True)

    def to_smt2(self):
        return "(ite {} {} {})".format(self.cond_clause.to_smt2(), self.then_clause.to_smt2(), self.else_clause.to_smt2())

class Implies(AST):
    def __init__(self, left, right):
        assert isinstance(left, AST)
        assert isinstance(right, AST)
        self.left = left
        self.right = right

    def __repr__(self):
        return "{}({}, {})".format(self.__class__.__name__, self.left, self.right)

    def getVariables(self):
        return self.left.getVariables() | self.right.getVariables()

    def getConditionVariables(self, is_parent_condition=False):
        return self.left.getConditionVariables(True)

    def to_smt2(self):
        return "(=> {} {})".format(self.left.to_smt2(), self.right.to_smt2())


from ...Exceptions import ExecutionError
from .helper import split_script_to_lines

from .Fnode import *
from .Command import *
from .SymbolType import SymbolType

class VegaSmtLibParser:
    def __init__(self):
        self.symbols = {}

    def parse_fnode_arg(self, last):
        last = last.strip(' ')
        # print("parse_fnode_arg({})".format(last)) # DEBUG
        if last[0] != '(':
            if ' ' in last:
                arg, last = last.split(' ', 1)
                return arg, last.strip(' ')
            else:
                return last, ""
        else:
            bracket_depth = 0
            arg = ""
            for i, c in enumerate(last):
                if c == '\r' or c == '\n':
                    continue
                arg += c
                if c == '(':
                    bracket_depth += 1
                elif c == ')':
                    bracket_depth -= 1
                    if bracket_depth == 0:
                        # print("arg='{}', last='{}'".format(arg, last[i+1:]))
                        return arg, last[i+1:].strip(' ')
        raise ExecutionError('Reached fail through')

    def parse_fnode(self, line):
        line = line.strip(' ')
        if line == "()":
            return Blank()
        if line[0] == '(' and line[-1] == ')':
            line = line[1:-1]
        # print("parse_fnode({})".format(line)) # DEBUG

        raw_command = line.split(' ', 1)
        
        if len(raw_command) == 1:
            if raw_command[0] in ['true', 'false']:
                return BoolConstant(raw_command[0])
            else:
                symbol_name = raw_command[0]
                symbol_type = self.symbols[symbol_name]
                return Symbol(symbol_name, symbol_type)
        else:
            name, last = raw_command

            args = []
            while last:
                arg, last = self.parse_fnode_arg(last)
                args.append(self.parse_fnode(arg))
            
            map_name_fnode = {
                'and': And,
                'or': Or,
                'not': Not,
                '=>': Implies,
                '=': Equals,
                'ite': Ite,
            }

            if name in map_name_fnode:
                return map_name_fnode[name](name, args)
            elif name == 'distinct':
                return Not(name, [Equals(name, args)])
            else:
                return Fnode(name, args)

    def parse_command(self, line):
        line = line.strip(' ')
        if line[0] == '(' and line[-1] == ')':
            line = line[1:-1]
        
        if ' ' in line:
            name, last = line.split(' ', 1)
            
            if name in ['set-info']:
                return Command(name, last.split(' '))
            
            elif name == 'declare-datatypes':
                last_replace_split = last.strip(' ').replace('(', '').replace(')', '').split(' ')
                datatype_name = last_replace_split[1]
                datatype_values = last_replace_split[2:]
                for value in datatype_values:
                    self.symbols[value] = SymbolType(datatype_name)
                return Command(name, [DataType(datatype_name, datatype_values)])
            
            elif name == 'declare-fun':
                last_replace_split = last.strip(' ').replace('(', '').replace(')', '').split(' ')
                symbol_name = last_replace_split[0]
                symbol_type = last_replace_split[2]
                self.symbols[symbol_name] = SymbolType(symbol_type)
                return Command(name, [])
            else:
                return Command(name, self.parse_fnode(line).args())
        
        else:
            name = line
            return Command(name, [])

    def get_script(self, raw_script):
        script = []
        for line in split_script_to_lines(raw_script):
            script.append(self.parse_command(line))
        return script


from .AST import *

### Check expr a and b are identical (not follows equality of a nor b)
def eq(a, b):
    assert isinstance(a, AST)
    assert isinstance(b, AST)
    return a == b

def is_expr(expr):
    return isinstance(expr, AST)


from functools import reduce
from six.moves import cStringIO

from ..Exceptions import *
from ..AST import *
from .pysmt.ExtendedSmtLibParser import ExtendedSmtLibParser
from .vega.VegaSmtLibParser import VegaSmtLibParser

def parse_fnode(fnode, sorts, depth=1):
    if fnode.is_and():
        res = []
        for x in fnode.args():
            res.append(parse_fnode(x, sorts))
        return And(*res)
        # return And(*reduce(lambda r, x: r.append(parse_fnode(x)), fnode.args(), []))
    
    elif fnode.is_or():
        res = []
        for x in fnode.args():
            res.append(parse_fnode(x, sorts))
        return Or(*res)
        # return Or(*reduce(lambda r, x: r.append(parse_fnode(x)), fnode.args(), []))
    
    elif fnode.is_not():
        return Not(parse_fnode(fnode.args()[0], sorts))
    
    elif fnode.is_implies():
        return Implies(
            parse_fnode(fnode.args()[0], sorts), 
            parse_fnode(fnode.args()[1], sorts)
            )
    
    elif fnode.is_symbol():
        symbol_name = fnode.symbol_name()
        sort_name = fnode.symbol_type().name
        if Value(symbol_name) in sorts[sort_name].values: # Check if symbol is member of datatype
            return Value(symbol_name)
        else:
            return Variable(symbol_name, sorts[sort_name])
    
    elif fnode.is_bool_constant():
        value = fnode.constant_value() # -> <class 'bool'>
        if value:
            return Top()
        else:
            return Bot()
    
    elif fnode.is_equals():
        return Eq(parse_fnode(fnode.args()[0], sorts), parse_fnode(fnode.args()[1], sorts))
    
    elif fnode.is_ite():
        return If(parse_fnode(fnode.args()[0], sorts), parse_fnode(fnode.args()[1], sorts), parse_fnode(fnode.args()[2], sorts))
    else:
        raise UnhandledCaseError("{}: node_type = {}".format(fnode, fnode.node_type()))

def parse_cmd(cmd, sorts):
    if cmd.name == 'declare-datatypes':
        values = [Value(x) for x in cmd.args[0].values]
        sorts[cmd.args[0].name] = Sort(cmd.args[0].name, values)
        return []
    
    if cmd.name == 'declare-fun':
        # symbol_name = cmd.args[0].symbol_name()
        # sort_name = cmd.args[0].symbol_type().name
        return []
    
    if cmd.name == 'assert':
        res = []
        for x in cmd.args:
            res.append(parse_fnode(x, sorts))
        return res
        # return reduce(lambda r, x: r.append(parse_fnode(x)), cmd.args, [])

### @public
def parse_and_get_script_from_file(file_name):
    with open(file_name) as f:
        return parse_and_get_script_from_file_stream(f)

### @public
def parse_and_get_script_from_file_stream(f):
    # parser = ExtendedSmtLibParser()
    # return parser.get_script(cStringIO(f.read()))

    parser = VegaSmtLibParser()
    return parser.get_script(f.read())

### @public
def parse_smt2_file(file_name):
    script = parse_and_get_script_from_file(file_name)

    ### Transform to expressions
    sorts = {}
    expr = []
    for cmd in script:
        res = parse_cmd(cmd, sorts) # NOTE: Do not replace `sorts` with `{}`. `sorts` passes referrence to `sorts`
        if res:
            expr += res

    ### Calcuate domain
    domain = set()
    for sort in sorts.values():
        domain |= sort.values

    return expr, Sort('Domain', domain)


class Command:
    def __init__(self, name, args):
        self.name = name
        self.args = args

    def __repr__(self):
        return "{}({}, {})".format(self.__class__.__name__, self.name, self.args)


from .Tactic import Tactic, WithReorder

class Feature:
    def __init__(self, debug=False, tactic=WithReorder()):
        assert isinstance(debug, bool)
        assert isinstance(tactic, Tactic)
        self.debug = debug
        self.tactic = tactic

        if self.debug: print("[*] vega.Feature: {}".format(self))

    def __repr__(self):
        return '{}(debug={}, tactic={})'.format(self.__class__.__name__, self.debug, self.tactic)

class FeatureCapability:
    def __init__(self, feature):
        assert isinstance(feature, Feature)
        self.feature = feature


# coding:utf-8
from . import AST
from .Map import RefMap
from .Satisfiability import Satisfiability, Sat

class Model(dict):
    def __init__(self, sat, variables, ref):
        assert isinstance(sat, Satisfiability)
        assert isinstance(ref, RefMap)
        assert isinstance(variables, dict) # Map Variable -> set を型にしたい
        self.sat = sat
        self.ref = ref
        self.variables = {} # ダンプを簡単にするために専用dictを用意して使いたい
        for key, value in variables.items():
            self.variables[key] = value

    def __repr__(self):
        if len(self.variables) < 100:
            return "{}(sat={}, {})".format(self.__class__.__name__, self.sat, self.variables)
        else: # Avoid too long output
            return "{}(sat={}, ...)".format(self.__class__.__name__, self.sat)
    
    def __getitem__(self, key):
        ref_key = self.ref.getRef(key)
        if ref_key in self.variables:
            return self.variables[ref_key]
        else:
            return None

    def __setitem__(self, key, value):
        # print("[*] Model::__setitem__(key={}, value={})".format(key, value)) # DEBUG
        assert isinstance(value, set) or isinstance(value, AST.Ref) # 専用の型を用意したい
        self.variables[self.ref.getRef(key)] = value

    def __bool__(self):
        return isinstance(self.sat, Sat)


class Satisfiability:
    def __bool__(self):
        return False

    def __eq__(a, b):
        return a.__class__.__name__ == b.__class__.__name__

    def __repr__(self):
        return "{}".format(self.__class__.__name__.lower())

class Sat(Satisfiability):
    def __and__(self, other):
        if isinstance(other, Sat) or isinstance(other, Unknown):
            return self
        return Unsat()

    def __or__(self, other):
        return self

    def __bool__(self):
        return True

class Unsat(Satisfiability):
    def __and__(self, other):
        return self

    def __or__(self, other):
        return other

class Unknown(Satisfiability):
    def __and__(self, other):
        return other

    def __or__(self, other):
        return other


import sys
import argparse

from .smtlib.app import evaluate_smt2_file

def usage(parser):
    parser.print_help(sys.stderr)
    sys.exit(1)

def main():
    version = "TODO" # TODO: load from setup.py or something else

    parser = argparse.ArgumentParser(description='vega [version {}] (c) Copyright 2020 TomoriNao (@K_atc).'.format(version))
    parser.add_argument("-smt2", action="store_true", default=True, help='use parser for SMT 2 input format (default: true)')
    parser.add_argument("-in", dest="_in", action="store_true", help='read formula from standard input')
    parser.add_argument("file", nargs="?")
    args = parser.parse_args()

    if args.smt2:
        if args._in:
            evaluate_smt2_file(sys.stdin)
        else:
            if args.file:
                with open(args.file) as f:
                    evaluate_smt2_file(f)
            else:
                print("[!] Specify `file`")
                usage(parser)

    


from .SymbolType import SymbolType

class Fnode:
    def __init__(self, name, args):
        self.name = name
        self._args = args

    def __repr__(self):
        return "{}({}, {})".format(self.__class__.__name__, self.name, args)

    def args(self):
        return self._args

    def node_type(self):
        return self.__class__.__name__

    def symbol_name(self):
        return None

    def is_and(self):
        return False
    
    def is_or(self):
        return False
  
    def is_not(self):
        return False
  
    def is_implies(self):
        return False
  
    def is_bool_constant(self):
        return False

    def is_symbol(self):
        return False

    def is_equals(self):
        return False

    def is_ite(self):
        return False

class Blank(Fnode):
    def __init__(self):
        Fnode.__init__(self)

class DataType(Fnode):
    def __init__(self, name, values):
        assert name
        self.name = name
        self.values = values

    def __repr__(self):
        return "{}(name={}, values={})".format(self.__class__.__name__, self.name, self.values)

class And(Fnode):
    def is_and(self):
        return True

class Or(Fnode):
    def is_or(self):
        return True

class Not(Fnode):
    def is_not(self):
        return True

class Implies(Fnode):
    def is_implies(self):
        return True

class BoolConstant(Fnode):
    def __init__(self, name):
        self.name = name

    def __repr__(self):
        return "{}({})".format(self.__class__.__name__, self.name)
    
    def constant_value(self):
        return bool(self.name)

    def is_bool_constant(self):
        return True

class Symbol(Fnode):
    def __init__(self, name, _type):
        assert isinstance(_type, SymbolType)
        self.name = name
        self.type = _type

    def __repr__(self):
        return "{}({})".format(self.__class__.__name__, self.name)

    def symbol_name(self):
        return self.name

    def symbol_type(self):
        return self.type

    def is_symbol(self):
        return True

class Equals(Fnode):
    def is_equals(self):
        return True

class Ite(Fnode):
    def is_ite(self):
        return True


def split_script_to_lines(raw_script):
    bracket_depth = 0
    lines = []
    line = ""
    comment_line = False
    for c in raw_script:
        if c == ';':
            comment_line = True
        elif c == '\r':
            pass
        elif c == '\n':
            line += ' '
            if comment_line:
                comment_line = False
                line = ""
        elif not comment_line:
            line += c
            if c == '(':
                bracket_depth += 1
            elif c == ')':
                bracket_depth -= 1
                if bracket_depth == 0:
                    lines.append(line)
                    line = ""
    return lines


class Tactic:
    def __repr__(self):
        return self.__class__.__name__

    def __eq__(a, b):
        return a.__class__.__name__ == b.__class__.__name__

### NOTE: Cannot solve constraints which needs re-order (e.g. P(y, x), Q(z), R(z, y) should be ordered in Q, R, P)
class Simple(Tactic):
    pass

### NOTE: Cannot use If statements in this tactic
class Simple2(Tactic):
    pass

class WithReorder(Tactic):
    pass


import sys
from functools import reduce

from .Exceptions import *
from . import AST
from .Law import applyDeMorgansLow
from .Map import RefMap
from .Model import Model
from . import Satisfiability
from .Feature import Feature, FeatureCapability
from . import Tactic
from .smtlib import SmtlibCapability

sat = Satisfiability.Sat()
unsat = Satisfiability.Unsat()
unknown = Satisfiability.Unknown()

class Solver(FeatureCapability, SmtlibCapability):
    def __init__(self, domain, feature=Feature()):
        assert isinstance(domain, AST.Sort)
        assert isinstance(feature, Feature)
        self.domain = domain
        FeatureCapability.__init__(self, feature)

        self.variables = {}
        self.ref = RefMap({})
        self.constraints = []
        self.post_constraints = []
        self.visited_variables = set()

        self.satisfiability = unknown # For optimization
        self.model_start = 0 # For optimization
        self.model_post_start = 0 # For optimization

        # print("Solver.ref = {}".format(self.ref)) # DEBUG
        assert not self.ref # Expect ref is blank

    def declareVariable(self, v):
        assert isinstance(v, AST.Variable)
        if not v in self.variables:
            self.variables[v] = set(self.domain.values)
            ## Exclude values not v.sort holds
            if not v.sort == self.domain:
                self.__addPostConstraint(AST.And(*[AST.Not(AST.Eq(v, t)) for t in self.domain.values - v.sort.values]))

    def add(self, *expr_list):
        for expr in expr_list:
            assert isinstance(expr, AST.AST), "expr={}".format(expr)
            self.constraints.append(expr)
            for v in expr.getVariables():
                self.declareVariable(v)
        self.satisfiability = unknown

    def __addPostConstraint(self, expr):
        assert isinstance(expr, AST.AST)
        self.post_constraints.append(expr)

    def check(self):
        if self.satisfiability == unknown:
            m = self.model()
            if m.sat:
                self.satisfiability = sat
            else:
                self.satisfiability = unsat
        return self.satisfiability

    ### @return Model
    def model(self):
        def __model_constraints(self, debug):
            for i, expr in enumerate(self.constraints[self.model_start:]): # Optimization
                try:
                    if debug and i > 0 and i % 100000 == 0: print("[*] Solver::model(): Solving constraint #{}".format(i)) # DEBUG
                    self.satisfiability = self.__evaluate(expr, self.__and_eq)
                    if debug: assert self.satisfiability == sat
                    if self.satisfiability == unsat:
                        break
                except Exception as e:
                    sys.stdout.flush()
                    print("[X] TheoremSolver::model::__model_constraints(): Exception occured in constraint #{}: {}".format(i, expr))
                    raise e
 
        def __model_post_constraints(self, debug):
            ### Algprithm for reordering problem (P(y, x), Q(z), R(z, y) should solved Q, R, P)
            ### But requires self.visited_variables to manage constrained variables (requied much memory?)
            ### @return sat/unsat
            def __with_reordering_constraints(self, debug):
                visited_variables = self.visited_variables
                unvisited_post_constraints = []
                for i, expr in enumerate(self.post_constraints):
                    try:
                        if not expr.getConditionVariables():
                            self.satisfiability = self.__evaluate_post(expr)
                            if self.satisfiability == unsat:
                                return
                            # visited_variables |= expr.getVariables()
                        else:
                            unvisited_post_constraints.append(expr)
                    except Exception as e:
                        sys.stdout.flush()
                        print("[X] TheoremSolver::model::__model_post_constraints(): Exception occured in post-constraint #{}: {}".format(i, expr))
                        raise e
                
                # print("unvisited_post_constraints = {}".format(unvisited_post_constraints)) # DEBUG
                skip_stop = len(unvisited_post_constraints)
                for i, expr in enumerate(unvisited_post_constraints):
                    try:
                        if debug and i > 0 and i % 100000 == 0: print("[*] Solver::model(): Solving unvisited post-constraint #{}".format(i)) # DEBUG
                        
                        cond_vars = set(map(lambda v: self.ref.getRef(v), expr.getConditionVariables()))
                        # print("cond_vars = {}".format(cond_vars)) # DEBUG
                        # print("visited_variables = {}".format(visited_variables)) # DEBUG

                        ### Give up determing vars in cond_vars from other constraints if i is bigger than length of original unvisited_post_constraints
                        if cond_vars & visited_variables == cond_vars or i >= skip_stop: # intersection
                            self.satisfiability = self.__evaluate_post(expr)
                            if self.satisfiability == unsat:
                                return
                            visited_variables |= set(map(lambda v: self.ref.getRef(v), expr.getVariables()))
                            skip_stop = len(unvisited_post_constraints)
                        else:
                            # print("unvisited_post_constraints: push {}".format(expr)) # DEBUG
                            unvisited_post_constraints.append(expr) # Push to last of iterator
                    except Exception as e:
                        print("[X] TheoremSolver::__model_post_constraints(): Exception occured in expression #{}: {}".format(i, expr))
                        raise e
                return self.satisfiability

            ### @return sat/unsat
            def __simple(self, debug):
                for i, expr in enumerate(self.post_constraints):
                    try:
                        if debug and i > 0 and i % 100000 == 0: print("[*] Solver::model(): Solving post-constraint #{}".format(i)) # DEBUG
                        self.satisfiability = self.__evaluate_post(expr)
                        if debug: assert self.satisfiability == sat
                        if self.satisfiability == unsat:
                            return unsat
                    except Exception as e:
                        print("[X] TheoremSolver::model(): Exception occured in post-constraint #{}: {}".format(i, expr))
                        raise e
                return self.satisfiability

            ### Without reordeing
            if isinstance(self.feature.tactic, Tactic.Simple):
                return __simple(self, debug)

            ### Iterate 2 times to solve reordering problem (but CANNOT support If statements)
            if isinstance(self.feature.tactic, Tactic.Simple2):
                res = __simple(self, debug)
                if res == unsat:
                    return unsat
                else:
                    return __simple(self, debug)
            
            ### With reordering
            if isinstance(self.feature.tactic, Tactic.WithReorder):
                return __with_reordering_constraints(self, debug)
            
            raise UnhandledCaseError("Unhandled tactic = {}".format(self.feature.tactic))
        
        debug = self.feature.debug

        if self.satisfiability == unknown: # Optimization
            ### Process constraints
            __model_constraints(self, debug)

            if self.satisfiability == sat:
                self.model_start = len(self.constraints) # Optimization

                ### NG: produces infinite loop
                # for expr in self.post_constraints:
                #     if expr.getConditionVariables():
                #         unvisited_variables = expr.getVariables() - expr.getConditionVariables()
                #         self.visited_variables -= set(map(lambda v: self.ref.getRef(v), unvisited_variables))

                ### Process post-constraints
                __model_post_constraints(self, debug)

                # if self.satisfiability == sat:
                #     self.model_post_start = len(self.post_constraints) # Optimization
            else:
                if debug: assert False

        return Model(self.satisfiability, self.variables, self.ref)

    def push(self):
        raise ExecutionError('Implement me')

    def pop(self):
        raise ExecutionError('Implement me')
        self.satisfiability = unknown

    def dumpConstraint(self):
        return '\n'.join([str(x) for x in self.constraints])

    def dumpPostConstraint(self):
        return '\n'.join([str(x) for x in self.post_constraints])

    def dumpRefMap(self):
        return self.ref.dump()

    def dumpVariables(self):
        return '\n'.join(['{} = {}'.format(k, v) for k, v in self.variables.items()])

    ### @return: sat or unsat
    def __and_eq(self, left, right):
        # print("[*] __and_eq(left={}, right={})".format(left, right)) # DEBUG
        # print("[*] self.ref.getRef(left) = {}".format(self.ref.getRef(left))) # DEBUG
        assert isinstance(left, AST.Variable)
        assert isinstance(right, AST.Value) or isinstance(right, AST.Variable)
        
        if isinstance(right, AST.Value):
            ref_left = self.ref.getRef(left)
            variables_ref_left = self.variables[ref_left]
            if isinstance(variables_ref_left, set) and not right in variables_ref_left: # Check intersection is not blank
                ### e.g. And(x == {Int}, y == x, y == {Pointer}) is unsat
                # print("[*] __and_eq(left={}, right={}): unsat".format(left, right)) # DEBUG
                # print("[*] __and_eq: variables = {}".format(self.variables)) # DEBUG
                if self.feature.debug: print("[!] __and_eq: symvar Ref({}) = {} cannot be {}".format(left, ref_left, right))
                return unsat
            self.variables[ref_left] = set([right])
            return sat
        
        elif isinstance(right, AST.Variable):
            if left != right: ### NOTE: Dismiss identical assign (i.e. x == x)
                # print("[*] self.variables[left] = AST.Ref(right): left={} right={}".format(left, right)) # DEBUG

                ref_left = self.ref.setRef(left, right)
                variables_left = self.variables[left]

                ### Handle lazy binding (i.e. late y == x)
                if isinstance(variables_left, set) and isinstance(self.variables[ref_left], set):
                    ### i.e. {y |-> {a}} |- {y == a} ~> {x |-> {a}} |- {y == a, y == x}
                    self.variables[ref_left] &= variables_left # intersection

                self.variables[left] = AST.Ref(right)

                if not self.variables[ref_left]: # blank
                    return unsat

            return sat

    ### @return: sat or unsat
    def __or_eq(self, left, right):
        assert isinstance(left, AST.Variable), "type(left)={}".format(type(left))
        assert isinstance(right, AST.Value), "type(right)={}".format(type(right))
        
        self.variables[self.ref.getRef(left)].add(right)
        return sat

    ### @return: sat or unsat
    def __not_eq(self, left, right):
        assert isinstance(left, AST.Variable)
        assert isinstance(right, AST.Value)
        
        ### NOTE: set::remove(key) emits KeyError if set contains key
        ref_left = self.ref.getRef(left)
        variables_ref_left = self.variables[ref_left]
        assert isinstance(variables_ref_left, set), "self.variables[{}] = {}".format(ref_left, variables_ref_left)
        
        if right in variables_ref_left:
            # self.variables[ref_left] &= ref_left.sort.values - set([right]) # intersection
            # self.variables[ref_left] &= self.domain.values - set([right]) # intersection
            self.variables[ref_left].remove(right)
        
        if self.variables[ref_left]:
            return sat
        else:
            if self.feature.debug: raise UnsatException("self.variables[{}] is blank".format(ref_left))
            return unsat

    ### Oneshot side-effect less satisfiability check
    def __check(self, expr):
        if isinstance(expr, AST.Top):
            return sat

        if isinstance(expr, AST.Bot):
            return unsat

        if isinstance(expr, AST.Eq):
            if isinstance(expr.v2, AST.Variable):
                if self.ref.getRef(expr.v1) == self.ref.getRef(expr.v2):
                    return sat
                else:
                    return unsat
            elif isinstance(expr.v2, AST.Value):
                ref_v1 = self.ref.getRef(expr.v1)
                variables_ref_v1 = self.variables[ref_v1]
                assert isinstance(variables_ref_v1, set), "self.variables[{}] = {}".format(ref_v1, variables_ref_v1)
                if expr.v2 in variables_ref_v1:
                    return sat
                else:
                    return unsat
        
        if isinstance(expr, AST.Not):
            if isinstance(expr.v1, AST.Eq):
                if self.__check(expr.v1) == sat:
                    return unsat
                else:
                    return sat

        if isinstance(expr, AST.And):
            return reduce(lambda r, x: r and self.__check(x), expr.v, sat)

        if isinstance(expr, AST.Or):
            return reduce(lambda r, x: r or self.__check(x), expr.v, unsat)
        
        raise UnhandledCaseError("expr: {}".format(expr))

    def __if(self, expr):
        assert isinstance(expr, AST.If)
        
        ### Flip cond clause expression (Remove Not() from cond clause)
        if isinstance(expr.cond_clause, AST.Not):
            return self.__if(AST.If(expr.cond_clause.v1, expr.else_clause, expr.then_clause))

        if self.__check(expr.cond_clause):
            return self.__if_then(expr)
        else:
            return self.__if_else(expr)
        
    def __if_then(self, expr):
        assert isinstance(expr, AST.If)
        res = sat
        ### Assume cond is true
        res &= self.__evaluate(expr.cond_clause, self.__and_eq)
        if self.feature.debug: assert res == sat # DEBUG
        res &= self.__evaluate(expr.then_clause, self.__and_eq)
        if self.feature.debug: assert res == sat # DEBUG
        return res

    def __if_else(self, expr):
        assert isinstance(expr, AST.If)

        if isinstance(expr.else_clause, AST.Terminate):
            return sat

        res = sat
        ### Assume cond is false
        not_cond_expr = applyDeMorgansLow(AST.Not(expr.cond_clause))
        res &= self.__evaluate(not_cond_expr, self.__and_eq)
        if not res == sat: print("[!] Solver::__if_else(expr={}): Not(if-cond) is {}".format(expr, res)) # DEBUG
        res &= self.__evaluate(expr.else_clause, self.__and_eq)
        if self.feature.debug: assert res == sat # DEBUG
        return res

    def __implies(self, expr):
        assert isinstance(expr, AST.Implies)

        if self.__check(expr.left):
            return self.__if_then(AST.If(expr.left, expr.right, AST.Terminate()))
        else:
            return sat

    def __or_on_x(self, expr, x):
        ### Get old type set
        ref_x = self.ref.getRef(x)
        prev_set = self.variables[ref_x] # NOTE: And(x == a, Or(x == b, x == c})) => unsat
        if isinstance(prev_set, AST.Ref):
            # print("[!] prev_set is not type set (potential bug). Assume ref_x is not constrained: prev_set={}, ref.get({})={}".format(prev_set, x, ref_x)) # DEBUG
            prev_set = set(ref_x.sort.values) # Assume prev_set is not constrainted
        
        ### Create new type set
        self.variables[ref_x] = set() # Set blank to collect OR conditions on symvar ref_x
        for v in filter(lambda e: x in e.getVariables(), expr.v):
            res = self.__evaluate(v, self.__or_eq)
            if res == unsat:
                self.variables[ref_x] = prev_set # Restore state for debugging
                if self.feature.debug: raise UnsatException("expression {} produces unsat".format(v))
                return unsat
        
        ### Merge old set and new set
        current_set = self.variables[ref_x]
        update_set = prev_set & current_set # intersection
        # print("[*] Solver::__evaluate(expr={}): ref_x={}, update_set={}".format(expr, ref_x, update_set)) # DEBUG
        
        if update_set:
            self.variables[ref_x] = update_set
            return sat
        else: # Blank
            if self.feature.debug: print("[!] __or_on_x: symvar Ref({}) = {} cannot be any values (prev={}, current={})".format(x, ref_x, prev_set, current_set))
            self.variables[ref_x] = prev_set # Restore state for debugging
            if self.feature.debug: raise UnsatException("variables[{}] is blank".format(ref_x))
            return unsat

    def __eq(self, expr, func):
        if isinstance(expr.v1, AST.Variable):
            if isinstance(expr.v2, AST.Variable) or isinstance(expr.v2, AST.Value):
                return func(expr.v1, expr.v2)
        raise UnhandledCaseError("expr={}".format(expr))

    def __evaluate(self, expr, func):
        # print("[*] __evaluate(expr={}, func={})".format(expr, func.__name__)) # DEBUG
        
        if isinstance(expr, AST.Terminate):
            ### Do nothing
            return sat

        if isinstance(expr, AST.Top):
            return sat

        if isinstance(expr, AST.Bot):
            return unsat
        
        if isinstance(expr, AST.Eq):
            if isinstance(self.feature.tactic, Tactic.WithReorder): self.visited_variables |= set(map(lambda v: self.ref.getRef(v), expr.getVariables()))
            return self.__eq(expr, func)
        
        if isinstance(expr, AST.And):
            if isinstance(self.feature.tactic, Tactic.WithReorder): self.visited_variables |= set(map(lambda v: self.ref.getRef(v), expr.getVariables()))
            return reduce(lambda r, x: r & self.__evaluate(x, self.__and_eq), expr.v, sat)
    
        if isinstance(expr, AST.Or):
            if isinstance(self.feature.tactic, Tactic.WithReorder): self.visited_variables |= set(map(lambda v: self.ref.getRef(v), expr.getVariables()))
            return reduce(lambda r, x: r | self.__or_on_x(expr, x), expr.getVariables(), unsat)
        
        if isinstance(expr, AST.Not):
            if isinstance(self.feature.tactic, Tactic.WithReorder): self.visited_variables |= set(map(lambda v: self.ref.getRef(v), expr.getVariables()))
            if isinstance(expr.v1, AST.Eq): # Not(Eq())
                return self.__eq(expr.v1, self.__not_eq)
            if isinstance(expr.v1, AST.Not): # Not(Not())
                return self.__evaluate(expr.v1.v1, func)
            if isinstance(expr.v1, AST.And) or isinstance(expr.v1, AST.Or): # Not(And()) or Not(Or())
                return self.__evaluate(applyDeMorgansLow(expr), func)

        ### If, Implies are handled in __evaluate_post() (see add())
        if isinstance(expr, AST.If):
            if isinstance(self.feature.tactic, Tactic.Simple2):
                raise ExecutionError("If statement is not supported by tactic {}".format(self.feature.tactic))
            self.__addPostConstraint(expr)
            return sat
        if isinstance(expr, AST.Implies):
            self.__addPostConstraint(expr)
            return sat

        raise UnhandledCaseError('expr: {}'.format(expr))
    
    ### Post phase evaluation
    def __evaluate_post(self, expr):
        if isinstance(expr, AST.If):
            return self.__if(expr)
        if isinstance(expr, AST.Implies):
            return self.__implies(expr)
        return self.__evaluate(expr, self.__and_eq) # Eq() etc.


from ..AST import Eq, Not

class SmtlibCapability:

    ### serialize constraints with z3 compatibility
    def to_smt2(self):
        res = []
        res.append("; generated by vega")
        res.append("(set-info :status unknown)")

        ### Collect sorts
        sorts = set()
        for v in self.variables.keys():
            sorts.add(v.sort)

        ### Define domain
        res.append("(declare-datatypes () (({} {})))".format(self.domain.name, ' '.join(["({})".format(x) for x in self.domain.values])))

        ### Define sorts
        for sort in sorts: # TODO
            res.append("(declare-datatypes () (({} {})))".format(sort.name, ' '.join(["({})".format(x) for x in sort.values])))
        for v in self.domain.values: # reffers domain.values to avoid redefine symbols (e.g. a of sort s, a of sort t)
            res.append("(declare-fun {} () {})".format(v.name, self.domain.name))

        ### Define all variables with same domain (sort)
        for v in self.variables.keys():
            res.append("(declare-fun {} () {})".format(v.name, self.domain.name))
        
        for expr in self.constraints:
            res.append("(assert\n {})".format(expr.to_smt2()))

        ### Restrict values
        for v in self.variables.keys():
            if v.sort != self.domain: # e.g. sort FileContent
                for value in self.domain.values - v.sort.values:
                    res.append("(assert\n {})".format(Not(Eq(v, value)).to_smt2()))

        res.append("(check-sat)")
        
        return '\n'.join(res)


from .AST import *

def applyDeMorgansLow(expr):
    # print("[*] applyDeMorgansLow(expr={})".format(expr)) # DEBUG
    assert isinstance(expr, Not)
    assert isinstance(expr.v1, Eq) or isinstance(expr.v1, And) or isinstance(expr.v1, Or)

    if isinstance(expr.v1, Eq):
        return expr
    if isinstance(expr.v1, And):
        return Or(*[Not(e) for e in expr.v1.v])
    if isinstance(expr.v1, Or):
        return And(*[Not(e) for e in expr.v1.v])
    


from six.moves import cStringIO

from ..Exceptions import *
from .parse import parse_cmd, parse_and_get_script_from_file_stream
from ..Solver import Solver, sat, unsat, unknown
from ..AST import Sort

def calcuate_domain(sorts):
    domain = set()
    for sort in sorts.values():
        domain |= sort.values
    return Sort('Domain', domain)

def check_sat(expr, sorts):
    Domain = calcuate_domain(sorts)
    s = Solver(Domain)
    s.add(*expr)
    res = s.check()
    print(res)
    return res

def get_model(expr, sorts):
    Domain = calcuate_domain(sorts)
    s = Solver(Domain)
    s.add(*expr)
    m = s.model()
    if m.sat == sat:
        print("(model")
        for v in m.variables.keys():
            if len(m[v]) > 1:
                ans = "(or {})".format(' '.join([x.name for x in m[v]]))
            else:
                ans = list(m[v])[0].name
            print("  (define-fun {name} () {sort_name}\n    (as {ans} {sort_name}))".format(name=v, sort_name=v.sort.name, ans=ans))
        print(")")
    return m.sat

def evaluate_smt2_file(file):
    script = parse_and_get_script_from_file_stream(file)

    ### Transform to expressions
    sorts = {}
    expr = []
    sat = unknown
    for line_no, cmd in enumerate(script):
        if cmd.name in ['declare-datatypes', 'declare-fun', 'assert']:
            res = parse_cmd(cmd, sorts) # NOTE: Do not replace `sorts` with `{}`. `sorts` passes referrence to `sorts`
            if res:
                expr += res
        elif cmd.name in ['check-sat']:
            sat = check_sat(expr, sorts)
        elif cmd.name in ['get-model']:
            sat = get_model(expr, sorts)
        elif cmd.name in ['set-info']:
            pass
        else:
            raise UnhandledCaseError("cmd = {}".format(cmd))

        if sat == unsat:
            print('(error "command #{}: model is not available")'.format(line_no))
            exit(1)


from functools import reduce
from six.moves import cStringIO
from collections import namedtuple

from pysmt.smtlib.parser import SmtLibParser
from pysmt.smtlib.script import SmtLibCommand
from pysmt.typing import PySMTType

from ...Exceptions import *

DataType = namedtuple('DataType', ['name', 'values'])

class ExtendedSmtLibParser(SmtLibParser):
    def __init__(self, env=None, interactive=False):
        SmtLibParser.__init__(self, env, interactive)

        self.commands["declare-datatypes"] = self._cmd_declare_datatypes
        self.commands["declare-fun"] = self._cmd_declare_fun

    def _parse_declare_datatypes(self, tokens):
        closing_clause_depth = 0
        parsed_tokens = []
        while True:
            tk = tokens.consume()
            parsed_tokens.append(tk)
            if tk == '(':
                closing_clause_depth += 1
            elif tk == ')':
                closing_clause_depth -= 1
            if closing_clause_depth == -1:
                break
        return parsed_tokens

    def _cmd_declare_datatypes(self, current, tokens):
        expr = self._parse_declare_datatypes(tokens)
        expr = list(filter(lambda x: not x in ['(', ')'], expr))
        ### Register variable for `declare-fun`
        self.cache.bind(expr[0], self._get_var(expr[0], PySMTType(basename=expr[0])))
        # for x in expr:
        #     self.cache.bind(x, self._get_var(x, PySMTType(basename=x)))
        return SmtLibCommand(name="declare-datatypes", args=[DataType(expr[0], expr[1:])])

    ### Overwrite definition of _cmd_declare_fun()
    def _cmd_declare_fun(self, current, tokens):
        """(declare-fun <symbol> (<sort>*) <sort>)"""
        var = self.parse_atom(tokens, current)
        params = self.parse_params(tokens, current)
        typename = self.parse_type(tokens, current)
        self.consume_closing(tokens, current)

        if params:
            typename = self.env.type_manager.FunctionType(typename, params)

        ### Cast FNode to PySMTType
        ### This fixes bug of original `_cmd_declare_fun`)
        typename = PySMTType(basename=typename.symbol_name())

        v = self._get_var(var, typename)
        if v.symbol_type().is_function_type():
            self.cache.bind(var, \
                    functools.partial(self._function_call_helper, v))
        else:
            self.cache.bind(var, v)
        return SmtLibCommand(current, [v])


from setuptools import setup, find_packages


with open('README.rst') as f:
    readme = f.read()

with open('LICENSE') as f:
    license = f.read()

with open('requirements.txt') as f:
    install_requires = f.read().split('\n')

setup(
    name='vega-solver',
    version='0.1.0',
    description='All sat style yet another theorem solver',
    long_description=readme,
    long_description_content_type='text/x-rst',
    author='Tomori Nao (@K_atc)',
    author_email='shiftx1026@gmail.com',
    url='https://github.com/K-atc/vega-solver',
    license=license,
    packages=find_packages(exclude=('tests', 'docs')),

    install_requires=install_requires,

    entry_points={
        "console_scripts": [
            "vega = vega.app:main"
        ]
    },

    classifiers=[
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'License :: OSI Approved :: MIT License',
    ],
)


# coding:utf-8
# from functools import lru_cache

from . import AST

class VariableToVariableMap:
    def __init__(self, values={}):
        # print("[*] VariableToVariableMap::__init__(values={})".format(values)) # DEBUG
        assert isinstance(values, dict)
        self.values = values
    
    def __repr__(self):
        return "{}({})".format(self.__class__.__name__, self.values)

    ### @return AST.Variable
    def __getitem__(self, key):
        assert isinstance(key, AST.Variable)
        return self.values[key]

    def __setitem__(self, key, value):
        assert isinstance(key, AST.Variable), "key={}".format(key)
        assert isinstance(value, AST.Variable), "value={}".format(value)
        # assert key != value # Prohibit identical mapping (i.e. f(x) = x) # TOO SLOW
        self.values[key] = value

    def __delitem__(self, key):
        self.values.__delitem__(key)

    def get(self, key, default):
        assert isinstance(key, AST.Variable), "type(key) = {}".format(type(key))
        assert isinstance(default, AST.Variable), "type(default) = {}".format(type(default))
        return self.values.get(key, default) # use Python built-in dict.get()

    def __iter__(self):
        return self.values.__iter__()

    def __len__(self):
        return self.values.__len__()

    def __bool__(self):
        return bool(self.values)

    def dump(self):
        if len(self.values) > 0:
            return '\n'.join(['{} -> {}'.format(k, v) for k, v in self.values.items()])
        else:
            return '()'


class RefMap(VariableToVariableMap):
    ### @return AST.Variable
    # @lru_cache(maxsize=None) # Optimization
    def getRef(self, key, call_count=0):
        # assert isinstance(key, AST.Variable)
        if call_count >= 1000:
            raise Exception("Dropped in infinite loop at Ref({})".format(key))
        
        ref_key = self.get(key, key)
        if key != ref_key:
            res = self.getRef(ref_key, call_count + 1)
            self[ref_key] = res # Memorize result
            return res
        else:
            return key # Fixed point

    ### @return AST.Variable: Ref of child (= `ref[child]`) 
    ### Eq(x, y) |-> x -> y, where x = child and y = Ref(parent)
    def setRef(self, child, parent):
        # print("[*] RefMap::set(child={}, parent={})".format(child, parent)) # DEBUG
        assert isinstance(child, AST.Variable), "ref: {} -> {}".format(child, parent)
        assert isinstance(parent, AST.Variable), "ref: {} -> {}".format(child, parent)
        
        x = child                       # x
        y = self.getRef(parent)         # y = Ref(parent)
        z = self.getRef(x)              # x'

        # self.getRef.cache_clear()       # Optimization

        ### Assume x is reflective: Ref(x) = x
        if x == z:                      # Do not use too slow `x in self`
            assert x.name != "Reg_7fffe33df877_r10_1683674" or y.name != "File_10_833c" # DEBUG
            self[x] = y                 # x == y
            return y                    # return Ref(x)
        
        ### Redirect relation of (child, parent) since parent is refferenced another child.
        ### Do not check $y != z$ because it calls Variable.__eq__()) and slows execution.
        ### i.e. ∀(x,y,z), {x == y, x == z} ~> {y == z, x == z} (allows x == y or y == z)
        else:
            assert y.name != "Reg_7fffe33df877_r10_1683674" or z.name != "File_10_833c" # DEBUG
            self[y] = z                 # y == z
            return z                    # return Ref(x)
